### WOL Proxy script for VMware vSphere v1.0
### 1) This script is automatically powers on VMs on ESXi hosts managed by vCenter Server by continuously checking for broadcast of magic packets (Wake on LAN packets).
### 2) To run this script you have to install VMware PowerCLI (https://my.vmware.com/web/vmware/details?downloadGroup=PCLI550&productId=352).
### 3) Please specify the correct parameters for vCenter Server address, username and password in config.xml configuration file before running this script.
### 4) This script if provided under Creative Commons Attribution 3.0 (CC BY 3.0, http://creativecommons.org/licenses/by/3.0/).
### 5) For any comments and requests please send me a e-mail on: omnimod@gmail.com

#Path to config.xml configuration file.
$ConfigPath = [System.IO.Path]::GetDirectoryName($myInvocation.MyCommand.Definition) + "\config.xml"

[xml]$XMLConfig = Get-Content ($ConfigPath)

############# Read parameters from config file #############
$VCAddress = $XMLConfig.Configuration.VC.Server 				#IP address or DNS name for connection to vCenter Server.
$VCUser = $XMLConfig.Configuration.VC.User 						#Username for connection to vCenter Server.
$VCPassword = $XMLConfig.Configuration.VC.Password 				#Password for connection to vCenter Server.
$RefreshInterval = $XMLConfig.Configuration.VC.Interval 		#Minimum VM list refresh interval in minutes.
$WOLPort = $XMLConfig.Configuration.WOL.Port 					#UDP port number for Magic packet listener. Default port is 9. for some applications you have to change port to 7.
$LoggingMode = $XMLConfig.Configuration.Logging.Logmode 		#0 - disable all logs output, 1 - writes logs to file, 2 - writes logs on screen, 3 - writes logs both to file and to on screen.
$Logfolder = $XMLConfig.Configuration.Logging.Logfolder 		#Path to log folder, C:\Programdata\WoL Proxy\ is used by default.
############################################################

#Detect log file path to writes logs into.
function Get-LogFilePath {
	$Date = Get-Date
	return ($Logfolder + $Date.Year.ToString() + "-" + $Date.Month.ToString() + "-" + $Date.Day.ToString() + "-wolproxy-logs.txt")
}

function Write-Log ($Message) {
	if ($LoggingMode -eq 1 -or $LoggingMode -eq 3) {

		#Create log folder if it is not exists
		if (!(Test-Path -Path $Logfolder)) {
			New-Item -Path $Logfolder -ItemType Directory
		}
		
		#Write logs to file
		$LogFile = Get-LogFilePath
		((Get-Date).ToString() + " " + $Message) >> $LogFile
	}

	if ($LoggingMode -ge 2) {
		Write-Host ((Get-Date).ToString() + " " + $Message)
	}
}

#Get VM Array from vCenter, which contains VM name and MAC address.
function Get-VMArray {
	Write-Log "Refreshing VM list from vCenter."
	$VMArray = Get-VM | Get-NetworkAdapter | Select-Object Parent,@{Name="MacAddress"; Expression = {$_.MacAddress.Replace(":","")}}
	return $VMArray
}

function Connect-VC {
	################################################################
	if ($global:DefaultVIServer.IsConnected -eq $false -or $global:DefaultVIServer.IsConnected -eq $null) {
			$ConnectionStatus = Connect-VIServer -Server $VCAddress -User $VCUser -Password $VCPassword
	}
	
	#If connection to vCenter Server failed, end the script.
	if ($global:DefaultVIServer.IsConnected -eq $false -or $global:DefaultVIServer.IsConnected -eq $null) {
		Write-Log $Error[0].ToString()
		break
	}
}

function Get-MAC ($DatagramContent) {
	$Packet = ""
		
	#Convert Byte array to String.
	foreach ($Byte in $DatagramContent) {
		if ($Byte -le 15) {
			$Packet += ("0"+[Convert]::ToString($Byte, 16))
		}
		else {
			$Packet += [Convert]::ToString($Byte, 16)
		}
	}

	#Check for the right size of UDP datagram, should be 102 bytes or 204 characters.
	if ($Packet.Length -eq 204) {

		#Check for correct UDP datagram header
		if ($Packet -match '^ffffffffffff') {
			#Get MAC address from UDP datagram
			$MAC = $Packet.Substring(12,12)
			$Counter = [Regex]::Matches($Packet, $MAC)
				
			#Check if MAC address is repeated for 16 times
			if ($Counter.Count -eq 16) {
				return $MAC
			}
		}
	}
}

########Start of the script########
Add-PSSnapin VMware.VimAutomation.Core

#Connect to vCenter Server
Write-Log ("Connecting to vCenter server: '" + $VCAddress + "'.")
Connect-VC

#Get VM list which include VM names and MAC addresses and set timestamp for refresh interval
$TimeStamp = Get-Date
$VMs = Get-VMArray

#Create UDP endpoint
Write-Log  ("Creating UDP Listener on port '" + $WOLPort.ToString() + "'.")
$endpoint = new-object System.Net.IPEndPoint ([IPAddress]::Any,$WOLPort)
$udpclient = new-Object System.Net.Sockets.UdpClient $WOLPort

###Main script
do {
	#Check if needs to refresh VM list
	$CurrentTime = Get-Date
	if ($TimeStamp.AddMinutes($RefreshInterval) -le $CurrentTime) {
		Connect-VC
		$VMs = Get-VMArray
		$TimeStamp = $CurrentTime
	}

	try {
		$DatagramContent = $udpclient.Receive([ref]$endpoint)
	}
	catch {
		Write-log "Error getting UDP datagram."
		break
	}
	
	if ($DatagramContent -ne $null) {
		#Check if UDP contains magic packet and MAC address
		$MAC = Get-MAC $DatagramContent
		
		#Check if MAC address belongs to any VM in VM list
		if ($MAC -ne $null) {
			$VM = ($VMs | Where-Object {$_.MacAddress -eq $MAC})
					
			if ($VM -ne $null) {
				Write-Log ("Received Magic Packet for '" + $VM.Parent + "'.")

				#Try to get VM from VC
				Connect-VC
				$FoundVM = Get-VM -Name $VM.Parent

				#Check if VM isn't already Powered On
				if ($FoundVM -ne $null) {
					if ($FoundVM.PowerState -ne "PoweredOn") {
						Write-Log ("Power on '" + $VM.Parent + "' with MAC Address: '" + $MAC + "'.")
						Start-VM -VM $VM.Parent -RunAsync
					}
					else {
						Write-Log ("VM '" + $VM.Parent + "' is already Powered On.")
					}
				}
				else {
					Write-Log ("Cannot find VM with MAC address: '" + $MAC.ToString() + "'.")
				}
			}
		}
	}
}
while($true)